function sequence = preprocessing(seq, mode)
%% sequence = preprocessing(seq, mode)
%
% the possible parameters for 'mode' are:
% ---------------------------------------
% - 'median'   : performs a distance-adaptive median filter on the
%                distances
%
% - 'threshld' : applies a dynamic amplitude threshold rejecting invalid
%                distance measurements via its amplitude value
%
% - 'edgerm'   : remove false measurements arising in the case of edges
%
% - 'open'     : applies an opening on the binary image of the usage vector
%                holding the valid distance measurements
%
% - 'normals'  : computes for each point its normal using its
%                8-neighborhood
%
% - 'ngb'      : stores for each point the indices of its 8-neighborhood
%
%
% - 'standard' : performs the methods of 'median', 'threshld', 'open',
%                'edgerm'
% - 'all'      : performs all available methods

disp('Preprocessing ...');

if(nargin < 2   ) mode = 'standard'; end

%% determine the methods that should be performed:
med = 0;
thres = 0;
edge = 0;
open = 0;
norm = 0;
ngb = 0;
svd = 0;

switch mode
    case 'all'
        med = 1;
        thres = 1;
        edge = 1;
        open = 1;
        norm = 1;
        ngb = 1;
    case 'median'
        med = 1;
    case 'threshld'
        thres = 1;
    case 'edgerm'
        edge = 1;
    case 'open'
        open = 1;
    case 'normals'
        norm = 1;
        ngb = 1;
    case 'ngb'
        ngb = 1;
    case 'svd'
        thres = 1;
        edge = 1;
        open = 1;
        svd = 1;
        norm = 1;
    case 'standard'
        med = 1;
        thres = 1;
        edge = 1;
        open = 1;
    otherwise
        disp('Error: Unknown value for parameter "mode"');
        return;
end

indx = [];
sequence = seq;
w = sequence.camType.width;
h = sequence.camType.height;

if(med)
    sequence.frames = cell2mat(arrayfun(@(x)(median_filtering(x, sequence.camType)), sequence.frames, 'UniformOutput', false));
end

if(thres)
    sequence.frames = cell2mat(arrayfun(@(x)(thresholding(x, sequence.camType)), sequence.frames, 'UniformOutput', false));
end

if(edge)
    sequence.frames = cell2mat(arrayfun(@(x)(edgepoint_removing(x, sequence.camType)), sequence.frames, 'UniformOutput', false));
end

if(open)
    sequence.frames = cell2mat(arrayfun(@(x)(open_close_usage(x, sequence.camType, 'open')), sequence.frames, 'UniformOutput', false));
end
    
if(norm)
    sequence.frames = cell2mat(arrayfun(@(x)(compute_normals(x, sequence.camType)), sequence.frames, 'UniformOutput', false));
end

if(ngb)
    i = 1 : w*h;
    pos = reshape(i, w, [])';
    i = mat2cell(i, 1, ones(1, w*h));
    indx = cellfun(@(x)(getNgbIndx(x, w, h, 3, true, pos)), i, 'UniformOutput', false);
    
    sequence.frames = cell2mat(arrayfun(@(x)(setfield(x, 'indx8', indx)), sequence.frames, 'Uniformoutput', false));
end

if(svd)
    sequence.frames = cell2mat(arrayfun(@(x)(svd_enhancing(x, sequence.camType)), sequence.frames, 'UniformOutput', false));
end
    
%% perform selected preprocessing for each frame of the sequence:
% f=1;
% while(f<=length(sequence.frames))
%     
%     % find points with z-values == 0: Points would lie on the image plane
%     use = find(sequence.frames(f).points3D_org(3,:) ~= 0);
%     u = ismember(sequence.frames(f).usage, use);
%     sequence.frames(f).usage = sequence.frames(f).usage(u);
%     
%     % if no point of a frame will be used (bad frame) this frame is deleted:
%     if(length(sequence.frames(f).usage) == 0)
%         sequence.frames(f) = [];
%     else
%         f=f+1;
%     end
% 
% end